# Implements a HD44780 character LCD connected via PCF8574 on I2C
# Should work with ESP8266 and (tested) Pyboard 1.1
#
# This is a compactification of the code in https://github.com/dhylands/python_lcd
#    (put everything in one file, remove constants, etc....) resulting in less memory consumption.

from time import sleep_ms, sleep_us

class I2cLcd: # Implements a HD44780 character LCD connected via PCF8574 on I2C

    def __init__(self, i2c, i2c_addr, n_lin, n_col):
        self.i2c = i2c
        self.i2c_addr = i2c_addr
        self.n_lin = min(n_lin, 4)
        self.n_col = min(n_col, 40)
        self.cur_x = 0
        self.cur_y = 0
        self.impl_nl = False
        self.backl = 0x08

        self.i2c.writeto(self.i2c_addr, bytearray([0]))
        sleep_ms(20)   # Allow LCD time to powerup

        self.i2c.writeto(self.i2c_addr, bytearray([0x34])) # LCD_FUNCTION_RESET, send reset 3 times
        self.i2c.writeto(self.i2c_addr, bytearray([0x30]))
        sleep_ms(5)    # need to delay at least 4.1 msec
        self.i2c.writeto(self.i2c_addr, bytearray([0x34])) # LCD_FUNCTION_RESET
        self.i2c.writeto(self.i2c_addr, bytearray([0x30]))
        sleep_ms(1)
        self.i2c.writeto(self.i2c_addr, bytearray([0x34])) # LCD_FUNCTION_RESET
        self.i2c.writeto(self.i2c_addr, bytearray([0x30]))
        sleep_ms(1)

        self.i2c.writeto(self.i2c_addr, bytearray([0x24])) # LCD_FUNCTION, put LCD into 4 bit mode
        self.i2c.writeto(self.i2c_addr, bytearray([0x20]))
        sleep_ms(1)
        
        self.display_off()
        self.clear()
        self.wr(0x06) # LCD_ENTRY_MODE | LCD_ENTRY_INC
        self.hide_cursor()
        self.display_on()

        cmd = 0x20  # LCD_FUNCTION
        if n_lin > 1:
            cmd |= 0x08  # LCD_FUNCTION_2LINES
        self.wr(cmd)

    def clear(self):  # Clears the LCD display and moves the cursor to the top left
        self.wr(0x01)  # LCD_CLR
        self.wr(0x02)  # LCD_HOME
        self.cur_x = 0
        self.cur_y = 0

    def show_cursor(self):  # Causes the cursor to be made visible, no blink
        self.wr(0x0e)  # LCD_ON_CTRL | LCD_ON_DISPLAY | LCD_ON_CURSOR

    def hide_cursor(self):  # Causes the cursor to be hidden
        self.wr(0xc)  # LCD_ON_CTRL | LCD_ON_DISPLAY

    def blink_cursor(self):  # Turns on the cursor, and makes it blink
        self.wr(0x0f)  # LCD_ON_CTRL | LCD_ON_DISPLAY | LCD_ON_CURSOR | LCD_ON_BLINK

    def display_on(self):  # Turns on (i.e. unblanks) the LCD
        self.wr(0x0c)  # LCD_ON_CTRL | LCD_ON_DISPLAY

    def display_off(self):  # Turns off (i.e. blanks) the LCD
        self.wr(0x08)  # LCD_ON_CTRL

    def backlight_on(self):  # Turns the backlight on
        self.backl = 0x08
        self.i2c.writeto(self.i2c_addr, bytearray([0x08]))

    def backlight_off(self):  # Turns the backlight off
        self.backl = 0x00
        self.i2c.writeto(self.i2c_addr, bytearray([0]))

    def move_to(self, cur_x, cur_y):  # Moves the cursor position to the indicated position
                                      # The cursor position is zero based (x == 0 -> first column)
        self.cur_x = cur_x
        self.cur_y = cur_y
        addr = cur_x & 0x3f
        if cur_y & 1:     # Lines 1 & 3 add 0x40
            addr += 0x40 
        if cur_y & 2:    # Lines 2 & 3 add number of columns
            addr += self.n_col
        self.wr(0x80 | addr) # LCD_DDRAM | ..

    def putchar(self, char):  # Writes character at the current cursor pos, advances the cursor by one pos
        if char == '\n':
            if self.impl_nl:  # self.impl_nl means we advanced due to a wraparound,
                pass          # so if we get a newline right after that we ignore it.
            else:
                self.cur_x = self.n_col
        else:
            self.wr(ord(char), 1)
            self.cur_x += 1
        if self.cur_x >= self.n_col:
            self.cur_x = 0
            self.cur_y += 1
            self.impl_nl = (char != '\n')
        if self.cur_y >= self.n_lin:
            self.cur_y = 0
        self.move_to(self.cur_x, self.cur_y)

    def putstr(self, string):  # Writes the string at the current cursor pos and advances cursor
        for char in string:
            self.putchar(char)

    def custom_char(self, location, charmap):   # Write a character to one of the 8 CGRAM locations,
                                                # available as chr(0) through chr(7).
        location &= 0x7
        self.wr(0x40 | (location << 3)) # LCD_CGRAM | ..
        sleep_us(40)
        for i in range(8):
            self.wr(charmap[i], 1)
            sleep_us(40)
        self.move_to(self.cur_x, self.cur_y)

    def wr(self, data, dbit=0):  # Write to the LCD; dbit: 0..command, 1..data
        byte = dbit | self.backl | data & 0xf0
        self.i2c.writeto(self.i2c_addr, bytearray([byte | 0x04]))
        self.i2c.writeto(self.i2c_addr, bytearray([byte]))
        byte = dbit | self.backl | ((data & 0x0f) << 4)
        self.i2c.writeto(self.i2c_addr, bytearray([byte | 0x04]))
        self.i2c.writeto(self.i2c_addr, bytearray([byte]))
        if not dbit and data <= 3: # The home and clear commands require a worst case delay of 4.1 msec
            sleep_ms(5)
